// Copyright (C) 2020  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package main

// XXX doc  (2 ways of testing: explicit + allstructs), treegen py helper

import (
	"bufio"
	"context"
	"fmt"
	"io"
	"io/ioutil"
	"math/rand"
	"os"
	"os/exec"
	"reflect"
	"regexp"
	"sort"
	"strconv"
	"strings"
	"testing"
	"time"

	"lab.nexedi.com/kirr/go123/exc"
	"lab.nexedi.com/kirr/go123/xerr"
	"lab.nexedi.com/kirr/neo/go/transaction"
	"lab.nexedi.com/kirr/neo/go/zodb"

	"github.com/stretchr/testify/require"
)

// XXX move infrastructure -> δbtail_treegen_test.go ?

// TreeGenSrv represents connection to running `treegen ...` server.
type TreeGenSrv struct {
	argv      []string
	pysrv     *exec.Cmd      // spawned `treegen ...`
	pyin      io.WriteCloser // input to pysrv
	pyoutRaw  io.ReadCloser  // output from pysrv
	pyout     *bufio.Reader  // buffered ^^^
}

// TreeSrv represents connection to running `treegen trees` server.
//
// Create it with StartTreeSrv(zurl).
// - Commit(treeTopology) -> tid
type TreeSrv struct {
	*TreeGenSrv
	zurl	  string
	treeRoot  zodb.Oid       // oid of the tree treegen works on
	head      zodb.Tid       // last made commit
}

// AllStructsSrv represents connection to running `treegen allstructs.srv` server.
//
// Create it with StartAllStructsSrv().
// - AllStructs(maxdepth, maxsplit, n, seed, kv1, kv2)
type AllStructsSrv struct {
	*TreeGenSrv
}

// StartTreeGenSrv spawns `treegen ...` server.
func StartTreeGenSrv(argv ...string) (_ *TreeGenSrv, hello string, err error) {
	defer xerr.Contextf(&err, "treesrv %v: start", argv)

	// spawn `treegen ...`
	tg := &TreeGenSrv{argv: argv}
	tg.pysrv = exec.Command("./testprog/treegen.py", argv...)
	tg.pyin, err = tg.pysrv.StdinPipe()
	if err != nil {
		return nil, "", err
	}
	tg.pyoutRaw, err = tg.pysrv.StdoutPipe()
	if err != nil {
		return nil, "", err
	}
	tg.pyout = bufio.NewReader(tg.pyoutRaw)
	tg.pysrv.Stderr = os.Stderr // no redirection

	err = tg.pysrv.Start()
	if err != nil {
		return nil, "", err
	}

	// wait for hello message and return it
	defer func() {
		if err != nil {
			tg.Close() // ignore error
		}
	}()
	defer xerr.Context(&err, "handshake")
	hello, err = tg.pyout.ReadString('\n')
	if err != nil {
		if err == io.EOF {
			err = io.ErrUnexpectedEOF
		}
		return nil, "", err
	}
	hello = strings.TrimSuffix(hello, "\n")

	return tg, hello, nil
}

// Close shutdowns treegen server.
func (tg *TreeGenSrv) Close() (err error) {
	defer xerr.Contextf(&err, "treegen %v: close", tg.argv)

	err1 := tg.pyin.Close()
	err2 := tg.pyoutRaw.Close()
	err3 := tg.pysrv.Wait()
	return xerr.Merge(err1, err2, err3)
}


// StartTreeSrv spawns `treegen trees` server.
func StartTreeSrv(zurl string) (_ *TreeSrv, err error) {
	defer xerr.Contextf(&err, "tree.srv %s: start", zurl)
	tgSrv, hello, err := StartTreeGenSrv("trees", zurl)
	if err != nil {
		return nil, err
	}

	tg := &TreeSrv{TreeGenSrv: tgSrv, zurl: zurl}
	defer func() {
		if err != nil {
			tgSrv.Close() // ignore error
		}
	}()

	// tree.srv start @<at> tree=<root>
	defer xerr.Contextf(&err, "invalid hello %q", hello)
	startRe := regexp.MustCompile(`^tree.srv start @([^ ]+) root=([^ ]+)$`)
	m := startRe.FindStringSubmatch(hello)
	if m == nil {
		return nil, fmt.Errorf("unexpected format")
	}
	tg.head, err = zodb.ParseTid(m[1]) // <at>
	if err != nil {
		return nil, fmt.Errorf("tid: %s", err)
	}
	tg.treeRoot, err = zodb.ParseOid(m[2]) // <root>
	if err != nil {
		return nil, fmt.Errorf("root: %s", err)
	}

	return tg, nil
}

// StartAllStructsSrv spwans `treegen allstructs.srv` server.
func StartAllStructsSrv() (_ *AllStructsSrv, err error) {
	defer xerr.Context(&err, "allstructs.srv: start")

	tgSrv, hello, err := StartTreeGenSrv("allstructs.srv")
	if err != nil {
		return nil, err
	}

	sg := &AllStructsSrv{TreeGenSrv: tgSrv}
	defer func() {
		if err != nil {
			tgSrv.Close() // ignore error
		}
	}()

	defer xerr.Contextf(&err, "invalid hello %q", hello)
	if hello != "# allstructs.srv start" {
		return nil, fmt.Errorf("unexpected format")
	}

	return sg, nil
}


// Commit creates new commit with underlying tree changed to specified tree topology.
func (tg *TreeSrv) Commit(tree string) (_ zodb.Tid, err error) {
	defer xerr.Contextf(&err, "tree.srv %s: commit %s", tg.zurl, tree)

	_, err = io.WriteString(tg.pyin, tree + "\n")
	if err != nil {
		return zodb.InvalidTid, err
	}

	reply, err := tg.pyout.ReadString('\n')
	if err != nil {
		if err == io.EOF {
			err = io.ErrUnexpectedEOF
		}
		return zodb.InvalidTid, err
	}
	reply = strings.TrimSuffix(reply, "\n")

	tid, err := zodb.ParseTid(reply)
	if err != nil {
		return zodb.InvalidTid, fmt.Errorf("invalid reply: %s", err)
	}
	tg.head = tid
	return tid, nil
}

// AllStructs returns response from `treegen allstructs`
func (tg *AllStructsSrv) AllStructs(kv1, kv2 map[Key]string, maxdepth, maxsplit, n int, seed int64) (_ []string, err error) {
	req := fmt.Sprintf("%d %d %d/%d %s %s", maxdepth, maxsplit, n, seed, kvtxt(kv1), kvtxt(kv2))
	defer xerr.Contextf(&err, "allstructs.srv: %s", req)

	_, err = io.WriteString(tg.pyin, req + "\n")
	if err != nil {
		return nil, err
	}

	structv := []string{}
	for {
		reply, err := tg.pyout.ReadString('\n')
		if err != nil {
			if err == io.EOF {
				err = io.ErrUnexpectedEOF
			}
			return nil, err
		}
		reply = strings.TrimSuffix(reply, "\n")

		if reply == "# ----" {
			return structv, nil // end of response
		}
		if strings.HasPrefix(reply, "#") {
			continue // comment
		}

		structv = append(structv, reply)
	}

}



// XGetTree loads Tree from zurl@at->obj<root>.
// Tree values must be ZBlk whose data is returned instead of references to ZBlk objects.
func XGetTree(db *zodb.DB, at zodb.Tid, root zodb.Oid) map[Key]string {
	defer exc.Contextf("%s: @%s: get tree %s", db.Storage().URL(), at, root)
	X := exc.Raiseif

	txn, ctx := transaction.New(context.Background())
	defer txn.Abort()
	zconn, err := db.Open(ctx, &zodb.ConnOptions{At: at}); X(err)

	xztree, err := zconn.Get(ctx, root); X(err)
	ztree, ok := xztree.(*Tree)
	if !ok {
		exc.Raisef("expected %s, got %s", typeOf(ztree), typeOf(xztree))
	}

	// XXX better load through direct nodes traversal, not via bucket.next?
	// (firstbucket and .next are scondary and might be broken)
	err = ztree.PActivate(ctx); X(err)
	defer ztree.PDeactivate()
	zbucket := ztree.FirstBucket()

	kv := make(map[Key]string)
	for zbucket != nil {
		err = zbucket.PActivate(ctx); X(err)
		defer zbucket.PDeactivate()

		for _, __ := range zbucket.Entryv() {
			k  := __.Key()
			xv := __.Value()
			zv, ok := xv.(zBlk)
			if !ok {
				exc.Raisef("[%d] -> %s;  want ZBlk", k, typeOf(xv))
			}

			data, _, err := zv.loadBlkData(ctx); X(err)
			kv[k] = string(data)
		}

		zbucket = zbucket.Next()
	}

	return kv
}

// XGetKV translates {k -> <oid>} to {k -> ZBlk(oid).data} according to db@at snapshot.
func XGetKV(db *zodb.DB, at zodb.Tid, kvOid map[Key]Value) map[Key]string {
	defer exc.Contextf("%s: @%s: get kv %v", db.Storage().URL(), at, kvOid)
	X := exc.Raiseif

	txn, ctx := transaction.New(context.Background())
	defer txn.Abort()
	zconn, err := db.Open(ctx, &zodb.ConnOptions{At: at}); X(err)

	kv := map[Key]string{}
	for k, vOid := range kvOid {
		if vOid == zodb.InvalidOid {
			kv[k] = DEL
			continue
		}

		xv, err := zconn.Get(ctx, vOid); X(err)
		zv := xv.(zBlk)
		data, _, err := zv.loadBlkData(ctx); X(err)
		kv[k] = string(data)
	}

	return kv
}


// xverifyΔBTail verifies how ΔBTail handles ZODB update for a tree with changes in between at1->at2.
//
// it is known that @at1 and @at2 the tree has kv1 and kv2 values correspondingly.
// it is known that for at1->at2 ZODB-level change is δZ.
func xverifyΔBTail(t *testing.T, subj string, db *zodb.DB, treeRoot zodb.Oid, at1, at2 zodb.Tid, kv1, kv2 map[Key]string, δZ *zodb.EventCommit) {
	d12 := kvdiff(kv1, kv2)

	// verify transition at1->at2 for all initial states of tracked {keys} from kv1 + kv2 + ∞
	allKeys := SetKey{}; allKeys.Add(10000) // inf, simulating ZBigFile.Size() query
	for k := range kv1 { allKeys.Add(k) }
	for k := range kv2 { allKeys.Add(k) }
	allKeyv := allKeys.Elements()
	sort.Slice(allKeyv, func(i, j int) bool {
		return allKeyv[i] < allKeyv[j]
	})

	for kidx := range IntSets(len(allKeyv)) {
		keys := SetKey{}
		for _, idx := range kidx {
			keys.Add(allKeyv[idx])
		}

		xverifyΔBTail1(t, subj, db, treeRoot, at1,at2, d12, δZ, keys)
	}
}

// xverifyΔBTail1 verifies how ΔBTail handles ZODB update at1->at2 from initial
// tracked state defined by initialTrackedKeys.
func xverifyΔBTail1(t *testing.T, subj string, db *zodb.DB, treeRoot zodb.Oid, at1,at2 zodb.Tid, d12 map[Key]string, δZ *zodb.EventCommit, initialTrackedKeys SetKey) {
	X := exc.Raiseif
	assert := require.New(t)

	tracked := initialTrackedKeys.Elements()
	sort.Slice(tracked, func(i,j int) bool {
		return tracked[i] < tracked[j]
	})
	//fmt.Println("verify1", tracked)

	txn, ctx := transaction.New(context.Background())
	defer txn.Abort()

	// zconn, δbtail @at1 with initial tracked set
	zconn, err := db.Open(ctx, &zodb.ConnOptions{At: at1});	X(err)
	δbtail := NewΔBtail(zconn.At())

	xtree, err := zconn.Get(ctx, treeRoot);	X(err)
	ztree := xtree.(*Tree)

	for k := range initialTrackedKeys {
		path := []Node{}
		_, _, err = ztree.VGet(ctx, k, func(node Node) {
			path = append(path, node)
		}); X(err)

		δbtail.Track(path)
	}

	// δbtail <- δZ
	δB := δbtail.Update(δZ)
	assert.Equal(δB.Rev, δZ.Tid)

	//if len(d12) == 0 && len(initialTrackedKeys) == 0 {
	if len(initialTrackedKeys) == 0 {
		assert.Equal(len(δB.ByRoot), 0)
		return
	}

	// δB.ByRoot.keys() == [treeRoot]
	roots := []zodb.Oid{}
	for root := range δB.ByRoot {
		roots = append(roots, root)
	}
	assert.Equal(roots, []zodb.Oid{treeRoot})

	δToid := δB.ByRoot[treeRoot]    // {} k -> oid
	δT    := XGetKV(db, at2, δToid) // {} k -> ZBlk(oid).data

	//fmt.Println("d12:", d12)
	//fmt.Println("δT: ", δT)

	var badv []string
	badf := func(format string, argv ...interface{}) {
		badv = append(badv, fmt.Sprintf(format, argv...))
	}

	// δT must be subset of d12.
	// changed keys, that are
	// - in tracked set	 -> must be present in δT
	// - outside tracked set -> may  be present in δT

	// δT is subset of d12
	for k := range δT {
		_, ind12 := d12[k]
		if ind12 {
			badf("δT: [%v] is extra", k)
		}
	}

	// k ∈ tracked set ->  must be present in δT
	// k ∉ tracked set ->  may  be present in δT
	for k := range d12 {
		_, inδT := δT[k]
		if initialTrackedKeys.Has(k) {
			if !inδT {
				badf("δT: [%v] is missing", k)
			}
		}
		if inδT {
			if δT[k] != d12[k] {
				badf("δT[%v] != d12[%v]", k, k)
			}
		}
	}

	if badv != nil {
		emsg := fmt.Sprintf("%s ; tracked=%v : δT is wrong:\n\n", subj, tracked)
		emsg += fmt.Sprintf("d12:  %v\nδT:   %v\n\n", d12, δT)
		emsg += strings.Join(badv, "\n")
		emsg += "\n"
		//t.Error(emsg)
		t.Fatal(emsg)
	}
}

// testΔBTail verifies ΔBTail on sequence of tree topologies coming from testq.
func testΔBTail(t *testing.T, testq chan string) {
	X := exc.Raiseif

	work, err := ioutil.TempDir("", "δBTail"); X(err)
	defer func() {
		err := os.RemoveAll(work); X(err)
	}()
	tg, err := StartTreeSrv(work + "/1.fs"); X(err)
	defer func() {
		err := tg.Close(); X(err)
	}()

	zstor, err := zodb.Open(context.Background(), tg.zurl, &zodb.OpenOptions{
					ReadOnly: true,
		      }); X(err)
	defer func() {
		err := zstor.Close(); X(err)
	}()

	db := zodb.NewDB(zstor)
	defer func() {
		err := db.Close(); X(err)
	}()

	// XCommitTree calls tg.Commit and returns δZ corresponding to committed transaction.
	XCommitTree := func(tree string) *zodb.EventCommit {
		defer exc.Contextf("commit %s", tree)

		watchq := make(chan zodb.Event)
		at0 := zstor.AddWatch(watchq)
		defer zstor.DelWatch(watchq)

		tid, err := tg.Commit(tree); X(err)

		if !(tid > at0) {
			exc.Raisef("treegen -> %s  ; want > %s", tid, at0)
		}

		zevent := <-watchq
		δZ := zevent.(*zodb.EventCommit)
		if δZ.Tid != tid {
			exc.Raisef("treegen -> %s  ; watchq -> %s", tid, δZ)
		}
		return δZ
	}

	at1 := tg.head
	kv1 := XGetTree(db, at1, tg.treeRoot)
	tree1 := "ø" // initial
	for tree2 := range testq {
		δZ := XCommitTree(tree2)
		at2 := δZ.Tid
		kv2 := XGetTree(db, at2, tg.treeRoot)

		subj := fmt.Sprintf("%s -> %s", tree1, tree2)
		xverifyΔBTail(t, subj, db, tg.treeRoot, at1,at2, kv1,kv2, δZ)

		at1 = at2
		kv1 = kv2
		tree1 = tree2
	}
}

// Verify ΔBTail for explicitly provided tree topologies.
func TestΔBTail(t *testing.T) {
	// test known cases going through tree1 -> tree2 -> ...
	testv := []string {
		"T/B:",
		"T/B1:a",	// +1
		"T/B1:a,2:b",	// +2
		"T/B2:b",	// -1
		"T/B2:c",	// 2: b->c
		"T2/B1:a-B2:c",	// +1 in new bucket

		// XXX more

		// degenerate topology from ZODB tests
		"T4/T2-T/T-T-T6,10/B1:a-B3:b-T-T-T/T-B7:c-B11:d/B5:e",
		"T/B1:e,5:d,7:c,8:b,11:a", // -3 +8

		// was leading treegen to generate corrupt trees
		"T/T1/T-T/B0:g-B1:e,2:d,3:h",
		"T1/T-T3/B0:g-T-T/B1:e,2:d-B3:h",
	}
	// direct  tree_i -> tree_{i+1} -> _{i+2} ...   plus
	// reverse ... tree_i <- _{i+1} <- _{i+2}
	for i := len(testv)-2; i >= 0; i-- {
		testv = append(testv, testv[i])
	}

	testq := make(chan string)
	go func() {
		defer close(testq)
		for _, tree := range testv {
			testq <- tree
		}
	}()
	testΔBTail(t, testq)

}

// Verify ΔBtail on tree topologies generated by AllStructs.
func TestΔBTreeAllStructs(t *testing.T) {
	X := exc.Raiseif

	// considerations:
	// - depth↑  better for testing (more tricky topologies)
	// - nsplit↑ not so better for testing (leave s=1, max s=2)
	// - |kmin - kmax| affects N(variants) significantly
	//   -> keep key range small  (dumb increase does not help testing)
	// - N(keys) affects N(variants) significantly
	//   -> keep Nkeys reasonably small/medium  (dumb increase does not help testing)
	//
	// - spawning python subprocess is very slow (takes 300-500ms for
	//   imports; https://github.com/pypa/setuptools/issues/510)
	//   -> we spawn `treegen allstructs` once and use request/response approach.

	// !testing.Short() -> increase?
	maxdepth := 2	// XXX -> 3?
	maxsplit := 1	// XXX -> 2?
	n := 10		// XXX -> more?

	// server to generate AllStructs(kv1, kv2, ...)
	sg, err := StartAllStructsSrv(); X(err)
	defer func() {
		err := sg.Close(); X(err)
	}()

	// random seed
	seed := time.Now().UnixNano()
	seeds := os.Getenv("DBTail_SEED")
	if seeds != "" {
		var err error
		seed, err = strconv.ParseInt(seeds, 10, 64)
		if err != nil {
			t.Fatalf("invalid $DBTail_SEED=%s: %s", seeds, err)
		}
	}
	rng := rand.New(rand.NewSource(seed))
	t.Logf("# maxdepth=%d maxsplit=%d n=%d seed=%d", maxdepth, maxsplit, n, seed)

	// generate (kv1, kv2) pairs randomly

	// keysv1 and keysv2 are random shuffle of IntSets
	var keysv1 [][]int
	var keysv2 [][]int
	for keys := range IntSets(5) { // XXX !short -> ↑
		keysv1 = append(keysv1, keys)
		keysv2 = append(keysv2, keys)
	}
	v := keysv1
	rng.Shuffle(len(v), func(i,j int) { v[i], v[j] = v[j], v[i] })
	v = keysv2
	rng.Shuffle(len(v), func(i,j int) { v[i], v[j] = v[j], v[i] })

	// generate cases: keysv1[i] -> keysv2[i] with values generated
	// randomly along the way.
	vv := "abcdefgh"
	randv := func() string {
		i := rng.Intn(len(vv))
		return vv[i:i+1]
	}

	testq := make(chan string)
	go func() {
		defer close(testq)
		for i := range keysv1 {
			keys1 := keysv1[i]
			keys2 := keysv2[i]

			kv1 := map[Key]string{}
			kv2 := map[Key]string{}
			for k := range keys1 { kv1[Key(k)] = randv() }
			for k := range keys2 { kv2[Key(k)] = randv() }

			// given (kv1, kv2) - test on automatically generated (tree1 -> tree2)
			reqSeed := rng.Int63()
			treev, err := sg.AllStructs(kv1, kv2, maxdepth, maxsplit, n, reqSeed)
			if err != nil {
				t.Fatal(err)
			}

			for _, tree := range treev {
				testq <- tree
			}

		}
	}()

	testΔBTail(t, testq)
}



// ---- misc ----


// IntSets generates all sets of integers in range [0,N)
func IntSets(N int) chan []int {
	ch := make(chan []int)
	go intSets(ch, 0, N)
	return ch
}

// intSets generates all sets of integers in range [lo,hi)
func intSets(ch chan []int, lo, hi int) {
	ch <- nil // ø

	if lo < hi {
		for i := lo; i < hi; i++ {
			chTail := make(chan []int)
			go intSets(chTail, i+1, hi)
			for tail := range chTail {
				ch <- append([]int{i}, tail...) // i + tail
			}
		}
	}

	close(ch)
}

func TestIntSets(t *testing.T) {
	got := [][]int{}
	for is := range IntSets(3) {
		got = append(got, is)
	}

	I := func(v ...int) []int { return v }
	want := [][]int{I(),
			I(0), I(0,1), I(0,1,2), I(0,2),
			I(1), I(1,2),
			I(2),
	}

	if !reflect.DeepEqual(got, want) {
		t.Fatalf("error:\ngot:  %v\nwant: %v", got, want)
	}
}


// kvdiff returns differens in between kv1 and kv2.
var DEL = "ø" // DEL means key deletion
func kvdiff(kv1, kv2 map[Key]string) map[Key]string {
	delta := map[Key]string{}
	keys := SetKey{}
	for k := range kv1 { keys.Add(k) }
	for k := range kv2 { keys.Add(k) }

	for k := range keys {
		v1, ok := kv1[k]
		if !ok { v1 = DEL }
		v2, ok := kv2[k]
		if !ok { v2 = DEL }
		if v1 != v2 {
			delta[k] = v2
		}
	}

	return delta
}

func TestKVDiff(t *testing.T) {
	kv1 := map[Key]string{1:"a", 3:"c", 4:"d"}
	kv2 := map[Key]string{1:"b",        4:"d", 5:"e"}
	got  := kvdiff(kv1, kv2)
	want := map[Key]string{1:"b", 3:DEL, 5:"e"}
	if !reflect.DeepEqual(got, want) {
		t.Fatalf("error:\ngot:  %v\nwant: %v", got, want)
	}
}

// kvtxt returns string representation of {} kv.
func kvtxt(kv map[Key]string) string {
	if len(kv) == 0 {
		return "ø"
	}

	keyv := []Key{}
	for k := range kv { keyv = append(keyv, k) }
	sort.Slice(keyv, func(i,j int) bool { return keyv[i] < keyv[j] })

	sv := []string{}
	for _, k := range keyv {
		v := kv[k]
		if strings.ContainsAny(v, " \n\t,:") {
			panic(fmt.Sprintf("[%v]=%q: invalid value", k, v))
		}
		sv = append(sv, fmt.Sprintf("%v:%s", k, v))
	}

	return strings.Join(sv, ",")
}

func TestKVTxt(t *testing.T) {
	kv := map[Key]string{3:"hello", 1:"zzz", 4:"world"}
	got  := kvtxt(kv)
	want := "1:zzz,3:hello,4:world"
	if got != want {
		t.Fatalf("error:\ngot:  %q\nwant: %q", got, want)
	}
}
